
/*
 * Portable Object Compiler (c) 1997,98.  All Rights Reserved.
 * $Id: runarray.m,v 1.1 1998/10/14 14:38:08 stes Exp $
 */

/*
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include "runarray.h"
#include "txtattr.h"
#include "ordcltn.h"

@implementation RunArray

- check
{
  int i, n;
  int j, m;
  unsigned s, p;
  id attrib, attrs;
  unsigned q = 0;
  unsigned totalsize = 0;

  for (i = 0, n = [values size]; i < n; i++) {
    attrs = [values at:i];
    assert([attrs size] > 0);
    attrib = [attrs lastElement];
    s = [attrib size];
    p = [attrib position];
    assert(p == q);
    for (j = 0, m = [attrs size]; j < m; j++) {
      attrib = [attrs at:j];
      assert([attrib size] == s);
      assert([attrib position] == p);
    }
    q = p + s;
    totalsize += s;
  }

  assert(totalsize == [self size]);
  return self;
}

+ new
{
  return [[super new] values:[OrdCltn new]];
}

- free
{
  [values free];
  return [super free];
}


- values
{
  return values;
}

- values:v
{
  values = v;
  return self;
}

- at:(unsigned)i
{
  int j, n;
  unsigned p, s;
  id attrib, attrs;

  for (j = 0, n = [values size]; j < n; j++) {
    attrs = [values at:j];
    attrib = [attrs lastElement];
    p = [attrib position];
    s = [attrib size];
    if (p <= i && i < (p + s))
      return attrs;
  }
  return nil;
}

static void 
setpossize(id c, unsigned pos, unsigned size)
{
  int n = [c size];

  while (n--)
    [[c at:n] position:pos size:size];
}

- (unsigned) size
{
  id attrib = [[values lastElement] lastElement];

  return [attrib position] + [attrib size];
}

- setsize:(unsigned)s
{
  id v, e;
  unsigned p, n;

  if ((v = [values lastElement])) {
    e = [v lastElement];
    p = [e position];
    n = [e size];
  } else {
    p = 0;
    n = 0;
  }

  if (p + n == s)
    return self;		/* size is correct */

  if ((p + n) < s) {
    id attrib = [TextAttribute new];	/* to fill */

    [attrib position:p + n size:(s - p - n)];
    [values add:[[OrdCltn new] add:attrib]];
    assert([self size] == s);
  } else {
    while (p >= s) {
      [[[values removeLast] freeContents] free];	/* to shrink */
      v = [values lastElement];
      e = [v lastElement];
      p = [e position];
    }
    setpossize(v, p, s - p);
  }

  assert([self check]);
  return self;
}

- (unsigned) runLengthAt:(unsigned)i
{
  return [[[self at:i] lastElement] size];
}

static void 
splitvalues(id values, unsigned q)
{
  id e, v;
  unsigned p, s;
  int i, n = [values size];

  for (i = 0; i < n; i++) {
    v = [values at:i];
    e = [v lastElement];
    p = [e position];
    s = [e size];
    if (p == q)
      return;			/* nothing to split */
    if ((p + s) > q) {
      [values at:i insert:[v deepCopy]];
      setpossize([values at:i], p, q - p);
      setpossize([values at:i + 1], q, p + s - q);
      return;
    }
  }
}

- addAttribute:attrib from:(unsigned)p size:(unsigned)s
{
  int i, j, n;

  if (s == 0)
    return self;

  splitvalues(values, p);
  splitvalues(values, (p + s));

  for (i = 0, n = [values size]; i < n; i++) {
    id v = [values at:i];
    id e = [v lastElement];
    unsigned q = [e position];

    if (p <= q && q < (p + s)) {
      unsigned es = [e size];

      assert((q + es) <= (p + s));
      [attrib reset];
      j = [v size];
      while (j--) {
	if ([attrib dominates:[v at:j]])
	  [[v removeAt:j] free];
      }
      if ([attrib set]) {
	[attrib position:q size:es];
	[v add:[attrib copy]];
      }
    }
  }

  [attrib free];
  assert([self check]);
  return self;
}

- coalesce
{
  int n = [values size];

  while (n--) {
    id v, w;

    if (!n)
      break;
    v = [values at:n];
    w = [values at:n - 1];
    if ([v isEqual:w]) {
      unsigned p = [[v lastElement] size];
      unsigned q = [[w lastElement] size];
      unsigned x = [[w lastElement] position];

      [[[values removeAt:n] freeContents] free];
      setpossize(w, x, p + q);
    }
  }

  return self;
}


- printOn:(IOD)aFile
{
  fprintf(aFile, "RunArray runs ");
  [values printOn:aFile];
  return self;
}

@end

