
/*
 * Portable Object Compiler (c) 1997,98.  All Rights Reserved.
 * $Id: blockxpr.m,v 1.17 1998/10/06 15:23:43 stes Exp $
 */

/*
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "blockxpr.h"
#include "output.h"
#include "options.h"
#include "module.h"
#include "identif.h"
#include "linetag.h"
#include "ocstring.h"
#include "compstat.h"
#include "ordcltn.h"
#include "token.h"
#include <assert.h>

id currentBlock;		/* inside genblockdefs compound */

@implementation BlockExpression
- block
{
  static int blockCount = 1;

  count = blockCount++;
  return self;
}

+ new
{
  return [[super new] block];
}

- free
{
  return nil;
}

- blkfuncname
{
  return [String sprintf:"blockFun%i", count];
}

- blkdataname
{
  return [String sprintf:"blockVar%i", count];
}

- blktypename
{
  return [String sprintf:"struct blockData%i", count];
}

- scope
{
  return scope;
}

- setArguments:aNode
{
  arguments = aNode;

  if (aNode) {			/* :x :y :z */
    int i;

    aNode = [aNode subnodes];
    argumentNames = [OrdCltn new];
    argumentTypes = [OrdCltn new];
    for (i = 1; i < [aNode size]; i += 2) {
      [argumentNames add:[aNode at:i]];
      [argumentTypes add:[Identifier str:"id"]];
    }
  }
  return self;
}

- setDecls:dNode body:bNode
{
  id lToken = [Token str:"{"];
  id rToken = [Token str:"}"];

  /* wrap statement in compound: handle inlinecaches and tempvars */
  compstmt =
      [CompoundStatement lbrace:lToken decls:dNode body:bNode rbrace:rToken];

  return self;
}

- setStatements:aNode
{
  return [self setDecls:nil body:aNode];
}

- setExpression:aNode
{
  id rNode = [Identifier str:"return "];
  id sNode = [Token str:";"];

  /* wrap return expression in compound */
  aNode = [[[[Node new] add:rNode] add:aNode] add:sNode];

  return [self setStatements:aNode];
}

- (int) nArgs
{
  if (argumentNames) {
    return [argumentNames size];
  } else {
    return 0;
  }
}

- grepTmpvars:aStmt
{

  return self;
}

- grepBlocks:aStmt
{
  [aStmt addBlock:self];
  /* do not descend, ie. don't get subblocks */
  return self;
}

- grepTopblks:aStmt
{
  [aStmt addTopblk:self];
  /* do not descend, ie. don't get subblocks */
  return self;
}

- grepCaches:aStmt
{
  return self;
}

- addScope:aScope
{
  if (!scope)
    scope = [OrdCltn new];
  [scope addIfAbsent:aScope];	/* no dups in -isSame sense */
  return self;
}

- blkdatatypOn:stream
{
  int n = [scope size];

  /* some compilers warn about empty structs */

  if (n) {
    int i;
    STR type = [[self blktypename] str];

    [stream echo:"%s {\n", type];
    for (i = 0; i < n; i++) {
      STR t, d;
      id s = [scope at:i];

      t = [[s scopetypename] str];
      d = [[s scopedataname] str];
      [stream echo:"%s *%s;\n", t, d];
    }
    [stream echo:"};\n"];
    [stream newline];
  }
  return self;
}

- packblockOn:stream
{
  /* wrap a blockData around the scope pointers */
  /* this is output-ing a comma expression */

  int n = [scope size];

  if (n) {
    int i;
    STR t = [[self blktypename] str];
    STR d = [[self blkdataname] str];

    [stream echo:"%s=(%s *)OC_Malloc(sizeof(%s)),\n", d, t, t];

    for (i = 0; i < n; i++) {
      STR ds, ts;
      id s = [scope at:i];

      ds = [[s scopedataname] str];
      ts = [[s scopetypename] str];
      [stream echo:"%s->%s=%s,", d, ds, ds];
    }
  }
  return self;
}

- unpackblockOn:stream
{
  /* sort of opposite of packblock */

  int n = [scope size];

  if (n) {
    int i;

    for (i = 0; i < n; i++) {
      STR ds, ts;
      id s = [scope at:i];

      ts = [[s scopetypename] str];
      ds = [[s scopedataname] str];
      [stream echo:"%s *%s= data->%s;\n", ts, ds, ds];
    }
  }
  return self;
}

- processBlocks
{
  /* now copy what's needed of the enclosing scope */
  /* this depends on the pushScope/popScope mechanism of compounds */

  [self addArgs:compstmt];
  [compstmt grepScope:self];

  /* descend for nested blocks */
  [Identifier pushScope:compstmt];
  [super processBlocks];
  [Identifier popScope];

  return self;
}

- pushScope
{
  int i, n = [scope size];

  for (i = 0; i < n; i++)
    [Identifier pushScope:[scope at:i]];
  return self;
}

- popScope
{
  int i, n = [scope size];

  for (i = 0; i < n; i++)
    [Identifier popScope];
  return self;
}

- addArgs:compound
{
  /* the :args before the '|' delimiter */
  /* treat them as initialized locals for subblocks */
  [compound addLocals:argumentNames types:argumentTypes];
  [compound addInitvars:argumentNames];
  return self;
}

- genblockdefsOn:stream
{
  STR tName;
  STR fnName = [[self blkfuncname] str];

  /* first descend into body for nested blocks */
  [compstmt genblockdefsOn:stream];

  [self blkdatatypOn:stream];
  [stream newline];

  /* tName is "void" if there is no data */
  if ([scope size]) {
    tName = [[self blktypename] str];
  } else {
    tName = "void";
  }

  [stream echo:"static id %s(id thisBlock,%s *data", fnName, tName];

  if (argumentNames) {
    int i;

    for (i = 0; i < [argumentNames size]; i++) {
      [stream echo:","];
      [[argumentTypes at:i] outputOn:stream];
      [stream echo:" "];
      [[argumentNames at:i] taglessOn:stream];
    }
  }
  [stream echo:")\n"];
  [stream echo:"{\n"];

  [self unpackblockOn:stream];

  [self pushScope];

  [stream echo:"if (1)\n"];

  currentBlock = self;

  [LineTag forceTag];
  [compstmt genOn:stream];

  currentBlock = nil;

  [self popScope];

  /* return by default - if not yet returned - the block itself */
  [stream newline];
  [stream echo:"return thisBlock;\n"];
  [stream echo:"}\n"];
  return self;
}

- genOn:stream
{
  if ([options blocks]) {
    int n = [self nArgs];
    int nData = [scope size];
    STR data, f = [[self blkfuncname] str];

    /* start comma expression */
    [stream echo:"("];

    if ([scope size]) {
      data = [[self blkdataname] str];
      [self packblockOn:stream];
    } else {
      data = "(void*)0";
    }

    [stream echo:"newBlock(%i,(IMP)%s,%s,%i))", n, f, data, nData];
    [LineTag forceTag];

    /* this is for automatic initialization */
    [Module usesEntry:[String str:"newBlock"]];
  } else {
    fatalError("blocks are not allowed with -noBlocks");
    [stream echo:"(id)0"];
  }

  return self;
}

@end

