
/*
 * Portable Object Compiler (c) 1997.  All Rights Reserved.
 * $Id: compstat.m,v 1.17 1998/10/06 15:23:43 stes Exp $
 */

/*
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "compstat.h"
#include "output.h"
#include "options.h"
#include "linetag.h"
#include "identif.h"
#include "set.h"
#include "ordcltn.h"
#include "ocstring.h"
#include "blockxpr.h"
#include "decls.h"
#include "methdecl.h"
#include <assert.h>

@implementation CompoundStatement
- grepTopblks:compound
{
  /* difference with grepBlocks: is that we override this */
  return self;
}

- lbrace:lToken decls:aDecls body:aBody rbrace:rToken
{
  static int theCount = 0;

  /* for generating unique types (blocks) */
  count = theCount++;

  /* using tokens for the braces, for nice lineTags */

  lBrace = lToken;
  [self add:lToken];

  decls = aDecls;
  if (aDecls) {
    [self add:aDecls];

    /* for local vars */
    [self setDeclarators:aDecls];

    /* descend to look for msg. expressions in initializers */
    [aDecls grepTmpvars:self];

    /* descend to look for message expressions for inlineCaches */
    if ([options inlineCache])
      [aDecls grepCaches:self];

    /* descend to look fo blocks inside this compound */
    [aDecls grepBlocks:self];
    [aDecls grepTopblks:self];
  }
  body = aBody;

  if (aBody) {
    [self add:aBody];

    /* descend to look for message expressions that need tempVars */
    [aBody grepTmpvars:self];

    /* descend to look for message expressions for inlineCaches */
    if ([options inlineCache])
      [aBody grepCaches:self];

    /* descend to look fo blocks inside this compound */
    [aBody grepBlocks:self];
    [aBody grepTopblks:self];
  }
  /* it is allowed to set rToken nil and include the rToken in the body */
  rBrace = rToken;
  if (rToken)
    [self add:rToken];

  return self;
}

+ lbrace:lToken body:aBody rbrace:rToken
{
  id newObj = [super new];

  [newObj lbrace:lToken decls:nil body:aBody rbrace:rToken];
  return newObj;
}

+ lbrace:lToken decls:aDecls body:aBody rbrace:rToken
{
  id newObj = [super new];

  [newObj lbrace:lToken decls:aDecls body:aBody rbrace:rToken];
  return newObj;
}


- setDeclarators:declarators
{
  localVariables = [declarators variables];
  localTypes = [declarators types];
  initvars = [declarators initvars];
  return self;
}

- blockVariables
{
  return blockVariables;
}

- addInitvars:vars
{
  if (!initvars)
    initvars = [Set new];
  [initvars addAll:vars];
  return self;
}

- addLocals:vars types:types
{
  if (!localVariables)
    localVariables = [OrdCltn new];
  if (!localTypes)
    localTypes = [OrdCltn new];

  /* this is to add self, _cmd to this compound */
  /* or to add arguments of a function to its compound statements */
  [localVariables addAll:vars];
  [localTypes addAll:types];

  return self;
}

- pushLocals
{
  if (localVariables) {
    [Identifier pushvars:localVariables types:localTypes];
  }
  return self;
}

- popLocals
{
  if (localVariables) {
    [Identifier popLocals:localVariables];
  }
  return self;
}

- refblockvar:anIdentifier
{
  id matching;

  if ((matching = [localVariables findMatching:anIdentifier])) {
    unsigned i = [localVariables offsetOf:matching];

    if (!blockVariables)
      blockVariables = [OrdCltn new];
    if (!blockTypes)
      blockTypes = [OrdCltn new];

    [blockTypes add:[localTypes removeAt:i]];
    [blockVariables add:[localVariables removeAt:i]];
    assert([blockTypes size] == [blockVariables size]);

    return self;		/* found identifier in this scope */
  } else {
    if ([blockVariables findMatching:anIdentifier]) {
      return self;		/* found identifier in this scope */
    }
  }

  return nil;			/* continue search */
}


- addTmpvar:aString
{
  if (!tmpVars)
    tmpVars = [OrdCltn new];
  [tmpVars add:aString];
  return self;
}

- addBlock:aBlock
{
  if (!blocks)
    blocks = [OrdCltn new];
  [blocks add:aBlock];
  return self;
}

- addTopblk:aBlock
{
  if (!myBlocks)
    myBlocks = [OrdCltn new];
  [myBlocks add:aBlock];
  return self;
}

- addCache:aString
{
  if (!inlineCaches)
    inlineCaches = [OrdCltn new];
  [inlineCaches add:aString];
  return self;
}

- grepTmpvars:aStmt
{
  return self;
}

- grepCaches:aStmt
{
  return self;
}

/* stes - may/2/97 - limit length of line output here */

#define NUM_VARS_PER_LINE 5

- printtmpvars:stream from:(int)i to:(int)j
{
  [stream echo:"id "];
  if (i < j)
    [stream echoId:[tmpVars at:i]];
  for (i++; i < j; i++) {
    [stream echo:","];
    [stream echoId:[tmpVars at:i]];
  }
  [stream echo:";\n"];
  return self;
}

- printtmpvars:stream
{
  int i, j;
  int n = [tmpVars size];

  assert(n > 0);

  for (i = 0; i < n; i = j) {
    j = i + NUM_VARS_PER_LINE;
    j = (j > n) ? n : j;
    [self printtmpvars:stream from:i to:j];
  }

  [LineTag forceTag];
  return self;
}


- printcaches:stream from:(int)i to:(int)j
{
  [stream echo:"static struct objcrt_inlineCache "];
  if (i < j)
    [stream echoId:[inlineCaches at:i]];
  for (i++; i < j; i++) {
    [stream echo:","];
    [stream echoId:[inlineCaches at:i]];
  }
  [stream echo:";\n"];
  return self;
}

- printcaches:stream
{
  int i, j;
  int n = [inlineCaches size];

  assert(n > 0);

  for (i = 0; i < n; i = j) {
    j = i + NUM_VARS_PER_LINE;
    j = (j > n) ? n : j;
    [self printcaches:stream from:i to:j];
  }

  [LineTag forceTag];
  return self;
}

- printblkdata:stream
{
  int i, n = [myBlocks size];

  for (i = 0; i < n; i++) {
    id aBlock = [myBlocks at:i];

    if ([[aBlock scope] size]) {
      STR type, var;

      type = [[aBlock blktypename] str];
      var = [[aBlock blkdataname] str];
      [stream echo:"%s *%s;\n", type, var];
    }
  }

  return self;
}


- scopetypename
{
  return [String sprintf:"struct scopeData%i", count];
}

- scopeconsname
{
  return [String sprintf:"newScopeData%i", count];
}

- scopedataname
{
  return [String sprintf:"scopeVar%i", count];
}

- scopedatatypeOn:stream
{
  int n = [blockVariables size];

  assert(n == [blockTypes size]);

  [stream newline];

  /* some compilers warn about empty structs */

  if (n) {
    int i;
    STR type = [[self scopetypename] str];

    [stream echo:"%s {\n", type];
    for (i = 0; i < n; i++) {
      [[blockTypes at:i] taglessOn:stream];
      [stream echo:" "];
      [[blockVariables at:i] taglessOn:stream];
      [stream echo:";\n"];
    }
    [stream echo:"};\n"];
  }
  return self;
}

- genscopeOn:stream
{
  /* no newline here */

  int n = [blockVariables size];

  assert(n == [blockTypes size]);

  if (n) {
    STR t = [[self scopetypename] str];
    STR d = [[self scopedataname] str];
    STR c = [[self scopeconsname] str];

    [stream echo:"%s *%s = %s();", t, d, c];
  }
  [LineTag forceTag];
  return self;
}

- copyscopeOn:stream
{
  int i, n = [blockVariables size];
  STR d = [[self scopedataname] str];

  assert(n == [blockTypes size]);

  /* only have to copy the value of vars that
   * have an initializer into the Block
   * (others are uninitialized anyway)
   */

  for (i = 0; i < n; i++) {
    id v = [blockVariables at:i];

    if ([initvars contains:v]) {
      [stream echo:"%s->", d];
      [v taglessOn:stream];
      [stream echo:" = "];
      [v taglessOn:stream];
      [stream echo:";"];
    }
  }

  if (n)
    [LineTag forceTag];
  return self;
}


- processBlocks
{
  if (blocks) {
    [Identifier pushScope:self];
    [super processBlocks];
    [Identifier popScope];
  }
  return self;
}

- scopeconsOn:stream
{
  int n = [blockVariables size];

  assert(n == [blockTypes size]);

  [stream newline];

  /* some compilers warn about unused funcs */

  if (n) {
    STR t = [[self scopetypename] str];
    STR c = [[self scopeconsname] str];

    [stream echo:"static %s *%s(void)\n", t, c];
    [stream echo:"{\n"];
    [stream echo:"\treturn (%s *)OC_Malloc(sizeof(%s));\n", t, t];
    [stream echo:"}\n"];
  }
  return self;
}

- genblockdefsOn:stream
{
  [self scopedatatypeOn:stream];
  [self scopeconsOn:stream];
  return [super genblockdefsOn:stream];
}

- genOn:stream
{
  [lBrace genOn:stream];

  [stream newline];

  /* generate temporary variables for message expressions */
  /* in this compound statement */
  if (tmpVars)
    [self printtmpvars:stream];

  /* generate static caches for the inline cache messenger */
  if (inlineCaches)
    [self printcaches:stream];

  /* declare block data if any blocks in this compound */
  if (blocks)
    [self printblkdata:stream];

  if (ismain) {
    /* generate a call to the runtime initializer
     * if (1) { body of main }
     */

    /* since this line has no equivalent in .m file, fake linetag
     * make it seem as if it's the line of the main() function itself
     */
    [outertag outputOn:stream];
    [stream echo:"objc_main_init();if (1) {\n"];
    [LineTag forceTag];
  }
  [self pushLocals];

  /* do this before decls so that id aBlock = [ ... ]; works */
  if ([blockVariables size]) {
    if (lBrace)
      [[lBrace lineTag] outputOn:stream];
    /* no newline here, otherwise have to emit a linetag */
    [self genscopeOn:stream];
  }
  /* translate decl so that initializer (e.g. id x = [...];)
   * works;  left-hand side is not translated to scope->x 
   * because pushScope
   * is done only AFTER dealing with these statement declarators
   * this also means that the rhs is only translated if the blockvar
   * is from some enclosing scope != this scope (that is sufficient)
   */

  if (decls) {
    [decls genOn:stream];
  }
  [Identifier pushScope:self];	/* after declarators ! */

  /* this needs to be done after decls
   * the right way to do this could perhaps be to eliminate 
   * declarations of variables that are accessed via scopePtr
   */
  if ([blockVariables size]) {
    /* no newline here, otherwise have to emit a linetag */
    [self copyscopeOn:stream];
  }
  /* make sure begin of statements have correct linetag */
  [stream newline];
  [LineTag forceTag];

  if (body) {
    [body genOn:stream];
  }
  [self popLocals];
  [Identifier popScope];

  [rBrace genOn:stream];


  if (ismain)
    [stream echo:"}"];		/* close the if(1) */

  return self;
}

- genmainOn:stream linetag:aTag
{
  ismain = YES;
  outertag = aTag;
  return [self genOn:stream];
}

@end

