
/*
 * Portable Object Compiler (c) 1997.  All Rights Reserved.
 * $Id: funcdef.m,v 1.16 1998/10/06 15:23:44 stes Exp $
 */

/*
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "funcdef.h"
#include "module.h"
#include "output.h"
#include "linetag.h"
#include "options.h"
#include "ocstring.h"
#include "identif.h"
#include "compstat.h"
#include <assert.h>

@implementation FunctionDefinition
- specifiers:aSpec declarators:aDecl extDeclarators:extDecl statements:aStmt
{
  specifiers = aSpec;
  if (aSpec)
    [self add:aSpec];		/* can be nil */

  declarators = aDecl;
  [self add:aDecl];
  extDeclarators = extDecl;
  if (extDecl)
    [self add:extDecl];
  statements = aStmt;
  [self add:aStmt];

  /* add arguments to compound (as if locals)
   * so that things will work if a Block references a function argument
   * it works for ANSI args and K&R style args, by sending a message
   * down the tree, which will be "caught" by the Declarators node that
   * hangs in there somewhere.
   */

  [declarators argsstmt:statements];
  if (extDecl)
    [extDecl argsstmt:statements];

  /* register function name for automatic runtime initialization
   */

  nameIdent = [declarators findIdentifier];
  name = [nameIdent string];
  if (!name)
    fatalError("function name not found");

  if ((aSpec != nil) && [aSpec containsStatic]) {
    [Module staticEntry:name];
  } else {
    [Module definesEntry:name];
  }

  return self;
}

+ specifiers:aSpec declarators:aDecl extDeclarators:extDecl statements:aStmt
{
  id newObj = [super new];

  [newObj specifiers:aSpec declarators:aDecl extDeclarators:extDecl statements:aStmt];
  return newObj;
}

- mainOn:stream
{
  STR fmt;
  id initCall = [options initCall];
  BOOL noShared = ([options sharedData] == NO);

  /* stes - may18/97 moved the xtern call to this place out of
     * function body for cplus
   */

  if (initCall) {
    STR f = [initCall str];

    if (noShared) {

      [stream echo:"extern struct useDescriptor *OCU_main;\n"];
      [stream echo:"extern struct modEntry *_objcModules;\n"];

      if ([options cplus]) {
	fmt = "extern \"C\" int %s(void *mods,void *desc);\n";
      } else {
	fmt = "extern int %s(void *mods,void *desc);\n";
      }

    } else {

      if ([options cplus]) {
	fmt = "extern \"C\" int %s(int debug,int traceInit);\n";
      } else {
	fmt = "extern int %s(int debug,int traceInit);\n";
      }

    }
    [stream echo:fmt, f];
  } else {
    fatalError("no initialization call defined");
  }

  if ([options filerCode]) {
    [stream echo:([options cplus]) ? "extern \"C\"" : "extern"];
    [stream echo:" void *_OBJCBIND_ascfiler(void);\n"];

    [stream echo:([options cplus]) ? "extern \"C\"" : "extern"];
    [stream echo:" void *_OBJCBIND_ocstring(void);\n"];

    /* this is only for the noShared, noPostlink case */
    /* so that autoinit will visit the AsciiFiler module */
    [Module usesEntry:[String str:"AsciiFiler"]];
    [Module usesEntry:[String str:"String"]];
  }
  /* stes - 6/97 for thread-safe runtime without recompiling runtime
   * stes - 3/98 removed the threads stuff and made user recompile
   * the runtime with threads, if user wants threads
   * same thing for the Boehm gc.  disadvantage is that it's a bit
   * more work for the user.
   */

  /* Seems if I define this static then program doesnt link for c++?
   */

  if (![options cplus])
    [stream echo:"static "];
  [stream echo:"int objc_main_init(void)\n"];
  [stream echo:"{\n"];

  if ([options filerCode]) {
    /* this is to pull in the AsciiFiler class */
    [stream echo:"_OBJCBIND_ascfiler();\n"];
    [stream echo:"_OBJCBIND_ocstring();\n"];
  }
  if (![options cache]) {
    [stream echo:"noCacheFlag=1;\n"];
  }
  if (![options nilRcvr]) {
    [stream echo:"noNilRcvr=1;\n"];
  }
  /* objcInit() call */
  if (noShared) {
    [stream echo:"%s(_objcModules,OCU_main);\n", [initCall str]];
  } else {
    [stream echo:"%s(0,0);\n", [initCall str]];
  }
  /* for auto-init */
  [Module usesEntry:initCall];

  [stream echo:"return 0;\n}\n\n"];
  return self;
}

/* override this method to generate a call to the initFunction */
/* of the Objective C runtime                                  */

- genOn:stream
{
  if ([name isEqual:[options mainFunction]]) {

    [self mainOn:stream];
    [LineTag forceTag];

    /* this is the entry point for objcInit() to intialize */
    /* all classes at startup */
    if (specifiers)
      [specifiers genOn:stream];
    [declarators genOn:stream];
    if (extDeclarators)
      [extDeclarators genOn:stream];
    [statements genmainOn:stream linetag:[nameIdent lineTag]];

    return self;
  } else {
    return [super genOn:stream];
  }
}

@end

