
/*
 * Portable Object Compiler (c) 1997.  All Rights Reserved.
 * $Id: identif.m,v 1.23 1998/10/06 15:23:44 stes Exp $
 */

/*
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "identif.h"
#include "ocstring.h"
#include "set.h"
#include "ordcltn.h"
#include "output.h"
#include "options.h"
#include "linetag.h"
#include "methimpl.h"
#include "methinfo.h"
#include "classdef.h"
#include "compstat.h"
#include "blockxpr.h"
#include <assert.h>

@implementation Identifier

/*****************************************************************************
 *
 * Creation
 *
 ****************************************************************************/


- string:aString
{
  lineTag = [LineTag currentTag];
  string = aString;
  return self;
}

- lineTag
{
  return lineTag;
}

+ str:(STR)s
{
  return [self string:[String str:s]];
}

+ string:aString
{
  return [[super new] string:aString];
}

- copy
{
  return [isa string:string];	/* get rid of lineTag */
}

- dontFree
{
  return self;
}

- free
{
  return nil;
}


/*****************************************************************************
 *
 * Equality (necessary for being member of Set)
 *
 ****************************************************************************/


- (int) asInt
{
  return [string asInt];	/* returns 0 for symbolic */
}

- string
{
  return string;
}

- (BOOL) isEqual:anIdentifier
{
  /* comparing nodes */
  if (isa != [anIdentifier class])
    return NO;

  if (self == anIdentifier) {
    return YES;
  } else {
    return [[self string] isEqual:[anIdentifier string]];
  }
}

- (unsigned) hash
{
  return [[self string] hash];
}

- (BOOL) isEqualSTR:(STR)s
{
  return [[self string] isEqualSTR:s];
}

- (BOOL) isVoid
{
  return [self isEqualSTR:"void"];
}

- (BOOL) isPtr
{
  return [self isEqualSTR:"STR"] || [self isEqualSTR:"id"];
}

- (STR) str
{
  return [[self string] str];
}


static id currlocals, curtypes;
static id currentScope;

- warnlocalivar
{
  if ([options warnlocalnst] && ([self isIvar] || [self isCvar])) {
    id msg;

    msg = [String sprintf:"Definition of '%s' hides instance variable", [self str]];
    warnError([msg str]);
  }
  return self;
}

+ pushScope:aScope
{
  if (!currentScope)
    currentScope = [OrdCltn new];
  [currentScope add:aScope];
  return self;
}

+ popScope
{
  return [currentScope removeLast];
}

+ pushvars:localVariables types:localTypes
{
  int i, n;

  if (!currlocals)
    currlocals = [OrdCltn new];
  if (!curtypes)
    curtypes = [OrdCltn new];

  assert([currlocals size] == [curtypes size]);

  for (i = 0, n = [localVariables size]; i < n; i++) {
    id v = [localVariables at:i];

    [v warnlocalivar];
    [currlocals add:v];
  }

  [curtypes addContentsOf:localTypes];
  assert([currlocals size] == [curtypes size]);

  return self;
}

+ popLocals:localVariables
{
  int n;

  n = (localVariables) ? [localVariables size] : 0;
  assert([currlocals size] == [curtypes size]);

  while (n--) {
    [currlocals removeLast];
    [curtypes removeLast];
  }

  assert([currlocals size] == [curtypes size]);
  return self;
}


- outputOn:stream
{
  [LineTag outputOn:stream until:lineTag];
  [stream echoId:string];
  return self;
}

- (BOOL) isSelf
{
  return [ClassDefinition inmethimpl] && [self isEqualSTR:"self"];
}

- (BOOL) isLocal
{
  return (currlocals != nil &&
	  [currlocals findMatching:self]);
}

- findblockvar
{
  /* "id aBlock = [ { aBlock } self ];"
   * lhs must not be considered
   * as blkvar inside statement declarators
   * this is achieved by pushing the scope only AFTER the stmt decls.
   * (assert noMark = NO).
   */

  int n = [currentScope size];

  while (n--) {
    id aScope = [currentScope at:n];

    if ([[aScope blockVariables] findMatching:self]) {
      assert(noMark == NO);
      return aScope;
    }
  }

  return nil;
}

- (BOOL) isIvar
{
  return [ClassDefinition iscurivar:self];
}

- (BOOL) isCvar
{
  return [ClassDefinition iscurcvar:self];
}

+ selfOn:stream
{
  id scope;
  static id aSelf;

  if (!aSelf)
    aSelf = [Identifier str:"self"];

  /*
   * stes - 10/97 in a Block, we need to access 'self' via the
   * block data.
   */

  if ((scope = [aSelf findblockvar])) {
    [aSelf genblockvarOn:scope on:stream];
  } else {
    [stream echoId:[aSelf string]];
  }

  return self;
}

+ selfidOn:stream
{
  [stream echo:"(id)("];
  [self selfOn:stream];
  [stream echo:")"];
  return self;
}

- genblockvarOn:aScope on:stream
{
  [stream echo:"%s->", [[aScope scopedataname] str]];
  [stream echoId:string];
  return self;
}

- genOn:stream ivarOf:aClassname
{
  if (![options otb]) {
    [Identifier selfOn:stream];
    [stream echo:"->"];
    [stream echoId:string];
  } else {
    [stream echo:"("];
    [Identifier selfOn:stream];
    [stream echo:"->ptr)->"];
    [stream echoId:string];
  }
  return self;
}

- genOn:stream cvarOf:aClass
{
  STR s = [aClass sharedname];

  /* maybe we could rather use a test to see whether 
   * self->isa is a factory or meta, at runtime
   * (would work for any "self" pointer)
   */

  if (![options otb]) {
    [stream echo:"((struct %s *)", s];
  } else {
    [stream echo:"((struct %s *)((struct OTB*)", s];
  }

  if ([currimpl isFactory]) {
    [Identifier selfOn:stream];
  } else {
    [stream echo:"("];
    [Identifier selfOn:stream];
    if (![options otb]) {
      [stream echo:"->isa)"];
    } else {
      [stream echo:"->ptr->isa)"];
    }
  }

  if (![options otb]) {
    [stream echo:")->", s];
  } else {
    [stream echo:")->ptr)->", s];
  }

  [stream echoId:string];
  return self;
}

- funheadOn:stream
{
  /* foo(x) translates to foo(x) if foo is ivar */
  /* (*foo)(x) translates to (*self->foo) if foo is ivar */

  [LineTag outputOn:stream until:lineTag];
  [stream echoId:string];
  return self;
}

- genfieldOn:stream
{
  /* foo->x translates to foo->x, foo.x to foo.x, if "x" is ivar */

  [LineTag outputOn:stream until:lineTag];
  [stream echoId:string];
  return self;
}
- genOn:stream
{
  id scope;

  [LineTag outputOn:stream until:lineTag];

  /* note that local vars take precedence over ivars.
   * also, local vars of a block take precedence over
   * blockdata variables (this works, since CompoundStatement
   * localvars of a Block do NOT contain the blockdata)
   * (unlike normal compounds inside compounds)
   */

  if ([self isSelf]) {
    [Identifier selfidOn:stream];	/* special cast rule */
  } else if ([self isLocal]) {
    [stream echoId:string];	/* use local, not ivar */
  } else if ([self isIvar]) {
    id aClass = [ClassDefinition currentName];

    [self genOn:stream ivarOf:aClass];
  } else if ([self isCvar]) {
    id aClass = [ClassDefinition current];

    [self genOn:stream cvarOf:aClass];
  } else if ((scope = [self findblockvar])) {
    [self genblockvarOn:scope on:stream];
  } else {
    if ([self isClassname] && [options sharedData] == NO) {
      [stream echo:"(*%s())", [self globalfun]];
    } else {
      [stream echoId:string];
    }
  }

  return self;
}

- taglessOn:stream
{
  [stream echo:" "];
  [stream echoId:string];
  return self;
}

- print
{
  [string printOn:stdout];
  return self;
}


- markdecl:(BOOL)flag
{
  noMark = flag;
  return self;
}

static id allbuiltinfuns;

- defbuiltinfun
{
  if (!allbuiltinfuns)
    allbuiltinfuns = [Set new];
  [allbuiltinfuns add:[self dontFree]];
  if ([options debugInfo]) {
    fprintf(stderr, "defined builtin function '%s'\n", [self str]);
  }
  return self;
}

static id templates;

- deftypename
{
  if (!allTypenames)
    allTypenames = [Set new];

  if (noMark) {
    [allTypenames add:[self dontFree]];
    if ([options debugInfo]) {
      fprintf(stderr, "defined typename '%s'\n", [self str]);
    }
  }
  return self;
}

- deftemplate
{
  /* hack for C++ templates, where type has limited scope */
  if (!templates)
    templates = [OrdCltn new];
  if (noMark)
    [templates add:[self dontFree]];
  return [self deftypename];
}

+ (BOOL) inTemplate
{
  return templates != nil;	/* see scanplus */
}

+ endTemplate
{
  if (templates) {
    [allTypenames removeContentsOf:templates];
    templates = nil;
    return self;
  } else {
    return self;
  }
}

- defclassname
{
  if (!allClassnames)
    allClassnames = [Set new];;
  [allClassnames add:[self dontFree]];
  return self;
}

- (BOOL) isbuiltinfun
{
  if (!allbuiltinfuns)
    allbuiltinfuns = [Set new];
  return [allbuiltinfuns contains:self];
}

- (BOOL) istypename
{
  if (!allTypenames)
    allTypenames = [Set new];
  return [allTypenames contains:self];
}

- (BOOL) isClassname
{
  if (!allClassnames)
    allClassnames = [Set new];
  return [allClassnames contains:self];
}

- (STR) globalfun
{
  id z = [String sprintf:"OBJCCLASS_%s", [self str]];

  return [z str];
}

- (STR) structfun
{
  id z = [String sprintf:"OBJCCFUNC_%s", [self str]];

  return [z str];
}

- (STR) superfun
{
  id z = [String sprintf:"OBJCCSUPER_%s", [self str]];

  return [z str];
}

- (STR) metastructfun
{
  id z = [String sprintf:"OBJCMFUNC_%s", [self str]];

  return [z str];
}

- (STR) metasuperfun
{
  id z = [String sprintf:"OBJCMSUPER_%s", [self str]];

  return [z str];
}

- findIdentifier
{
  return self;
}


- grepIvars:aSet addItype:bNode to:bSet
{
  /* because this can be sent to an initializer ( id o = foo(); )
   * we only want the identifier which was marked ('o')
   */

  if (noMark) {
    [aSet add:[self dontFree]];
    [bSet add:bNode];
  }
  return self;
}

- grepIvars:aSet initvars:iv addItype:bNode to:bSet
{
  [self grepIvars:aSet addItype:bNode to:bSet];
  return self;
}

- grepScope:aBlock
{
  int n = [currentScope size];

  while (n--) {
    id aScope = [currentScope at:n];

    if ([aScope refblockvar:self]) {
      [aBlock addScope:aScope];
      break;
    }
  }

  /* within a method implementation, can reference instance
   * variables.  those are used "by reference" (via self pointer)
   * we don't copy the ivar into the Block, but rather the self pointer.
   */

  if (currimpl) {
    /* stes 1/98 we can do this for every identifier
     * but have to be a bit careful about functions, methods etc.
     * with names that are ivars.  this is done by overriding
     * -grepScope: in other places -- not here
     */
    if ([self isIvar] || [self isCvar]) {
      [[Identifier str:"self"] grepScope:aBlock];
    }
  }
  return self;
}

- (BOOL) containsStatic
{
  return NO;
}

- forMethod:aMethod
{
  /* case of unary selector, no arguments */

  [aMethod addComponent:[self string]];
  return self;
}

@end

