
/*
 * Portable Object Compiler (c) 1997.  All Rights Reserved.
 * $Id: methdecl.m,v 1.13 1998/10/06 15:23:44 stes Exp $
 */

/*
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "methdecl.h"
#include "options.h"
#include "output.h"
#include "linetag.h"
#include "methinfo.h"
#include <assert.h>

@implementation MethodDeclarator

- free
{
  return nil;
}

- isFactory:(BOOL)flag
{
  isFactory = flag;
  return self;
}

- (BOOL) isFactory
{
  return isFactory;
}

- resType:aNode
{
  resultType = aNode;
  return self;
}

- setSelector:aNode
{
  selector = aNode;
  return self;
}

- setEllipsis:aToken
{
  ellipsis = aToken;
  return self;
}

- method:minfo
{
  /* to create MethodDecl from MethInfo */
  method = minfo;
  return self;
}


- processMethod
{
  id aMethod = [MethodInfo new];

  assert(selector);
  [selector forMethod:aMethod];
  if (ellipsis)
    [aMethod addEllipsis:ellipsis];
  if (resultType)
    [aMethod resType:resultType];
  [aMethod isFactory:isFactory];

  return aMethod;
}

- method
{
  if (!method)
    method = [self processMethod];
  return method;
}

- grepLocals:aCompound
{
  [method grepLocals:aCompound];
  return self;
}


- commentOn:stream
{
  [stream newline];
  [stream newline];
  [stream echo:"/*\n"];
  [stream echo:" * "];
  [self taglessOn:stream];
  [stream echo:"\n */\n"];
  [stream newline];
  return self;
}

- outputOn:stream
{
  [stream echo:(isFactory) ? "+" : "-"];
  return [super outputOn:stream];
}

- genOn:stream
{
  if ([options comments])
    [self commentOn:stream];

  /* force computation of method */
  [[self method] funprotoOn:stream];

  [LineTag forceTag];
  return self;
}

- grepScope:aMethod
{
  /* instead of just blindly processing nodes, we do it
   * a bit more selectively, by passing this to MethodInfo.
   * so that only arguments are actually processed.
   */
  return [method grepScope:aMethod];
}

@end

