
/*
 * Portable Object Compiler (c) 1997.  All Rights Reserved.
 * $Id: methimpl.m,v 1.15 1998/10/06 15:23:44 stes Exp $
 */

/*
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "methimpl.h"
#include "ocstring.h"
#include "output.h"
#include "options.h"
#include "classdef.h"
#include "linetag.h"
#include "ordcltn.h"
#include "methinfo.h"
#include "methdecl.h"
#include <assert.h>

id currimpl;

@implementation MethodImplementation

- free
{
  return [super free];
}

- (BOOL) isFactory
{
  id declarator = [subnodes at:0];

  return [declarator isFactory];
}

/* 
 * genblockdefsOn: is overridden so that the block can reference
 * instance variables
 *
 */

- processBlocks
{
  id declarator = [subnodes at:0];
  id compound = [subnodes at:1];
  id aClass = [ClassDefinition current];

  if (!aClass) {
    fatalError("method implementation outside class definition");
  }
  /* allow method arguments + self,_cmd to be referenced in Block */
  [declarator grepLocals:compound];

  /* allow iVar / cVar to be referenced */
  currimpl = self;
  [aClass makecurivars];

  [super processBlocks];

  [aClass clearcurivars];
  currimpl = nil;

  return self;
}

- genblockdefsOn:stream
{
  id aClass = [ClassDefinition current];

  [aClass startimpl:stream];

  /* remember which method implementation for "super" context */
  /* limit the scope of ivars to method implementations */

  currimpl = self;
  [aClass makecurivars];

  [super genblockdefsOn:stream];

  [aClass clearcurivars];
  currimpl = nil;

  return self;
}

- lineTag:aTag
{
  lineTag = aTag;
  return self;
}

- genOn:stream
{
  id declarator = [subnodes at:0];
  id compoundStmt = [subnodes at:1];
  id aClass = [ClassDefinition current];

  if (!aClass) {
    fatalError("method implementation outside class definition");
  }
  [aClass startimpl:stream];

  assert([declarator isKindOf:[MethodDeclarator class]]);

  [stream newline];
  [stream newline];

  /* this is so that all tags are flushed before we start
   * emitting purely syntetic code
   */
  if (lineTag)
    [LineTag outputOn:stream until:lineTag];

  /* remember which method implementation for "super" context */
  /* limit the scope of ivars to method implementations */
  currimpl = self;
  [declarator genOn:stream];
  [aClass makecurivars];
  [compoundStmt genOn:stream];
  [aClass clearcurivars];
  currimpl = nil;

  return self;
}

@end

