
/*
 * Portable Object Compiler (c) 1997.  All Rights Reserved.
 * $Id: module.m,v 1.58 1998/10/25 12:20:24 stes Exp $
 */

/*
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "module.h"
#include "ordcltn.h"
#include "output.h"
#include "options.h"
#include "ocstring.h"
#include "node.h"
#include "dictnary.h"
#include "set.h"
#include "linetag.h"
#include "classdef.h"
#include "methinfo.h"
#include "msgxpr.h"
#include <assert.h>
#include <string.h>		/* strtok */
#include <ctype.h>		/* isalnum */

@implementation Module

static void 
externc(id stream)
{
  /* stes - may 16, 1997 */
  if ([options cplus])
    [stream echo:"extern \"C\" "];
}

/*
 * definition of the Objective C module struct
 */

static char *modType = "\n"
"struct modDescriptor {\n"
/* keep this compatible with objcc */
"\tchar  *modName;\n"
"\tchar  *modVersion;\n"
"\tlong  modStatus;\n"
"\tchar  *modMinSel;\n"
"\tchar  *modMaxSel;\n"
"\tid    *modClsLst;\n"
"\tshort modSelRef;\n"
"\tchar  **modSelTbl;\n"
"\tstruct methodDescriptor *modMapTbl;\n"
/* extensions */
"};\n"
;

/*
 * Inline Cache
 * datatype must match objcrt.m !!
 */

+ cachemsgOn:stream
{
  STR expdir = [[options bind] str];

  [stream echo:"\n"];

  /* datatype must match objcrt.m !! */
  if ([options cplus]) {
    [stream echo:"struct objcrt_inlineCache {id cls;id (*imp)(...);};\n"];
  } else {
    [stream echo:"struct objcrt_inlineCache {id cls;id (*imp)();};\n"];
  }

  /* messages to nil with inlineCache */
  /* emit expdir so that the definition matches the one in objcrt */

  if ([options cplus]) {
    externc(stream);
    [stream echo:"id %s _nilHandler(...);\n", expdir];
  } else {
    [stream echo:"id %s _nilHandler(id,char*);\n", expdir];
  }

  return self;
}

+ checkbindOn:stream
{
  [stream echo:"\nextern char *objcrt_bindError(char *);\n"];
  return self;
}

+ otbvars:stream
{
  /* can add other fields here in the future */
  [stream echo:"\tstruct OTB *previnst;\n"];
  [stream echo:"\tstruct OTB *nextinst;\n"];
  return self;
}

+ prologueOn:stream
{
  STR e = [[options bind] str];
  STR bName = [[self objcbindName] str];

  if ([options comments])
    [stream echo:"\n/* objc prologue */\n\n"];

  /* Note that "id" is not defined in header files
     * but is actually a token for the compiler
     * allows for compatibility with stepstone and GNU
     * where "id" is also built in, and our def would clash
   */

  /* Change for Microsoft C : don't define "id" as a pointer to
     * 'struct _PRIVATE' (MSC doesn't do late defined structs)
     * We could define id as void* but let's use a generic def that
     * doesn't hide the "isa" pointer
     * At the same time, this requires a change for generation of
     * instance variable access (self->aVar) needs a struct _PRIVATE
     * cast, now.  This breaks something like anObject->isa, but that's
     * less important IMHO anyhow.
   */

  if ([options otb]) {
    [stream echo:"struct _PRIVATE {\n"];
    [stream echo:"\tstruct OTB *isa;\n"];
    [stream echo:"};\n"];
    [stream echo:"struct OTB {\n"];
    [stream echo:"\tstruct _PRIVATE *ptr;\n"];
    [Module otbvars:stream];
    [stream echo:"};\n"];
    [stream echo:"typedef struct OTB *id;\n"];
  } else {
    [stream echo:"struct _PRIVATE { struct _PRIVATE *isa; };\n"];
    [stream echo:"typedef struct _PRIVATE *id;\n"];
  }

  /* define C messenger */
  /* stes - apr 13, 97 - changed to prototyped definition */

  if (![options fwd]) {
    if ([options cplus]) {
      externc(stream);		/* in C++ IMP is defined as id *(...) */
      [stream echo:"id %s (* _imp(id,char*))(...);\n", e];
      externc(stream);
      [stream echo:"id %s (* _impSuper(id,char*))(...);\n", e];
    } else {
      [stream echo:"extern id %s (* _imp(id,char*))();\n", e];
      [stream echo:"extern id %s (* _impSuper(id,char*))();\n", e];
    }
  }
  /* define inline cache C messenger */

  if ([options inlineCache])
    [self cachemsgOn:stream];

  /* Table of pointers into __msgSelectors.
     * This table is modified by objcInit
     * is referenced in the source so needs prototype
     *
     * stes 22/7/1997 - force a reference to the OBJCBIND function.
     * This is to prevent some linkers (Metrowerks !) to
     * optimize BIND functions away when the module has no class definition
     * and only message expressions (which refer to selTransTbl).
     *
     * stes 2/1/98 - lcc4.0 was otherwise emitting a warning for this
     * assignment, so introduced -refBind to only do this for mwcc
   */

  [stream newline];
  [stream echo:"extern struct modDescriptor %s *%s(void);\n", e, bName];
  if ([options refBind]) {
    [stream echo:"static char **selTransTbl = (char **)%s;\n", bName];
  } else {
    [stream echo:"static char **selTransTbl;\n"];
  }

  /* stes 6/98 support for fwdimp() */

  if ([options fwd]) {
    if ([options cplus]) {
      [stream echo:"static id (**fwdTransTbl)(...);\n"];
    } else {
      [stream echo:"static id (**fwdTransTbl)();\n"];
    }
  }
  /* struct used in sharedType & defined by Stepstone objcc */
  if ([options cplus]) {
    [stream echo:"\nstruct _SLT {char *_cmd;id (*_imp)(...);};\n"];
  } else {
    [stream echo:"\nstruct _SLT {char *_cmd;id (*_imp)();};\n"];
  }

  /* Generate a type for Objective C modules 
     * make a forward define, for class impl
     * Microsoft C NEEDS this to come before sharedType (since
     * shared Type contains a struct modDescriptor pointer).
   */
  [stream echo:modType];
  [stream echo:"\nextern struct modDescriptor %s;\n",
   [[self moddescname] str]];

  /* The BIND function initializes the selTransTbl.  If it somehow
     * does not get initialized we can (at the cost of some test)
     * issue an error message.
   */
  if ([options checkbind])
    [self checkbindOn:stream];

  [LineTag forceTag];
  return self;
}

static id classimpl;
static id clsCltn;		/* if more than one allowed per file */

+ uniqclass:classImpl
{
  assert([options onePfile]);
  if (classimpl) {
    fatalError("only one implementation per file allowed");
  }
  classimpl = classImpl;
  return self;
}

+ addClsimpl:classImpl
{
  assert(![options onePfile]);
  if (!clsCltn)
    clsCltn = [OrdCltn new];
  [clsCltn add:classImpl];
  return self;
}

static id selCltn;

+ (int) selOffset:selname
{
  int n;
  id val;
  static id seldic;

  if (!selCltn) {
    selCltn = [OrdCltn new];
    seldic = [Dictionary new];
  }
  n = [selCltn size];

  if ((val = [seldic atKey:selname]))
    return [val asInt];

  [selCltn add:selname];
  [seldic atKey:selname put:[String sprintf:"%i", n]];

  return n;
}

static id fwdCltn;

+ (int) fwdOffset:msg
{
  int n;
  id val;
  static id msgdic;

  if (!fwdCltn) {
    fwdCltn = [OrdCltn new];
    msgdic = [Dictionary new];
  }
  n = [fwdCltn size];

  /* try to find another selector with same argument types */
  /* see msgxpr -hash and -isEqual: */
  /* it's okay if this fails, just will generate more code */

  if ((val = [msgdic atKey:msg]))
    return [val asInt];

  [fwdCltn add:[msg methodProto]];
  [msgdic atKey:[msg copy] put:[String sprintf:"%i", n]];

  return n;
}

/*****************************************************************************
 *
 * This is a (slightly) incompatible extension that allows us to have more
 * than one class implementation per file.
 *
 ****************************************************************************/

+ modclslstOn:stream
{
  int i, n;

  n = (clsCltn) ? [clsCltn size] : 0;

  [stream newline];
  [stream echo:"static id _modClsLst[] ={\n"];

  for (i = 0; i < n; i++) {
    STR s = [[[clsCltn at:i] idName] str];

    /* be careful : &_Foo is generated here */
    /* that is exactly the value of "id Foo := (id)&_Foo" */
    [stream echo:"(id)&_%s,\n", s];
  }

  /* must be NULL terminated */
  [stream echo:"(id)0};\n"];

  return self;
}

/***************************************************************************
 *
 * static char *_selTransTbl[] ={
 * "str:",
 * "printOn:",
 * 0};
 *
 **************************************************************************/

+ seltranstblOn:stream
{
  int i, n;

  n = (selCltn) ? [selCltn size] : 0;

  [stream newline];
  [stream echo:"static char *_selTransTbl[] ={\n"];

  for (i = 0; i < n; i++) {
    STR s = [[selCltn at:i] str];

    [stream echo:"\"%s\",\n", s];
  }

  /* always at least one entry (SGI cc doesn't like empty decls)    */
  [stream echo:"0\n};\n"];

  return self;
}

+ fwdstubsOn:stream
{
  int i, n;

  [stream newline];
  [stream newline];

  /* this can generate a lot of output
   * a good test in -fwdOffset: is important to reduce the size
   */

  n = (fwdCltn) ? [fwdCltn size] : 0;
  for (i = 0; i < n; i++) {
    id m = [fwdCltn at:i];

    [m argstructOn:stream];
    [m dispfunOn:stream];
    [m fwdstubOn:stream];
  }

  return self;
}

+ fwdtranstblOn:stream
{
  int i, n;

  n = (fwdCltn) ? [fwdCltn size] : 0;

  [stream newline];
  if ([options cplus]) {
    [stream echo:"static id (*(_fwdTransTbl[]))(...) ={\n"];
  } else {
    [stream echo:"static id (*(_fwdTransTbl[]))() ={\n"];
  }

  for (i = 0; i < n; i++) {
    STR s = [[fwdCltn at:i] fwdname];

    if ([options cplus]) {
      [stream echo:"(id(*)(...))%s,\n", s];
    } else {
      [stream echo:"(id(*)())%s,\n", s];
    }
  }

  /* always at least one entry (SGI cc doesn't like empty decls)    */
  if ([options cplus]) {
    [stream echo:"(id(*)(...))0\n};\n"];
  } else {
    [stream echo:"(id(*)())0\n};\n"];
  }

  return self;
}

/*****************************************************************************
 *
 * static struct modDescriptor _modDesc = {
 *         "testdir/test.m",
 *         "objc1.10.2",
 *
 ****************************************************************************/

#define MOD_MORETHANONE 0x4L	/* must match objcrt.m value !! */

+ moddescOn:stream
{
  STR modVersion = "objc1.10.2";
  int selSize = (selCltn) ? [selCltn size] : 0;
  STR modName = [[options srcfilename] str];
  STR modDescName = [[self moddescname] str];
  long modStatus = ([options onePfile]) ? 0 : MOD_MORETHANONE;

  [stream newline];

#if 0
  [stream echo:"static struct modDescriptor _modDesc = {\n"];
#endif
  /* removed static for problem with C++ forward "static" declarations */
  [stream echo:"struct modDescriptor %s = {\n", modDescName];
  [stream echo:"\t\"%s\",\n", modName];
  [stream echo:"\t\"%s\",\n", modVersion];
  [stream echo:"\t%iL,\n", modStatus];	/* modStatus */

/* min and max Sel are obsoleted, it seems */
  [stream echo:"\t0,\n"];	/* modMinSel */
  [stream echo:"\t0,\n"];	/* modMaxSel */
/*
   if (selSize) {
   [stream echo:"\t_selTransTbl[0],\n"];
   [stream echo:"\t_selTransTbl[%i],\n",selSize - 1];
   } else {
 */

  if ([options onePfile]) {
    /* this is the pre 1.4.x code with "one class per file" */
    if (classimpl) {
      STR cName = [[classimpl idName] str];

      [stream echo:"\t&%s,\n", cName];	/* modClsLst */
    } else {
      [stream echo:"\t0,\n"];	/* modClsLst */
    }
  } else {
    /* default since 1.4.x is to have more than one per file */
    [stream echo:"\t_modClsLst,\n"];	/* modClsLst */
  }

  if (selSize) {
    [stream echo:"\t%i,\n", selSize];	/* modSelRef */
    [stream echo:"\t_selTransTbl,\n"];	/* modSelTbl */
  } else {
    [stream echo:"\t0,\n"];	/* modSelRef */
    [stream echo:"\t0,\n"];	/* modSelTbl */
  }

  /* we don't support static references _mapTbl */
  /* this is just here for Stepstone compatibility */
  [stream echo:"\t0\n};\n"];	/* modMapTbl */

  return self;
}

/*****************************************************************************
 *
 * struct modDescriptor * _BIND_file()
 * 	{ return &_modDesc; }
 *
 ****************************************************************************/

/* This function is currently a UNIX-ism ...
 * Replaced strrchr (which seems non-standard on SunOS4.1.1)
 * Maybe should move this to String object.
 */

static char *
strRchr(const char *s, int c)
{
  char *t = (char *) s;

  while (*t)
    t++;
  while (t != s)
    if (*t == c)
      return t;
    else
      t--;
  return (*t == c) ? t : NULL;
}

static id 
ocbindname(const char *filename)
{
  id s;
  char *p, *pathSep = [options pathSeparator];

  p = strRchr(filename, pathSep[0]);
  s = [String str:(p) ? p + 1 : (char *) filename];
  p = [s strCopy];		/* get malloc'ed String buffer */
/* Original (doesnt work for foo-bar.m filename) :
   strtok(p,".");
   marcel@djo.nl replaced by block below: */
  {
    char *cp;

    if ((cp = strRchr(p, '.')) != NULL)
      *cp = '\0';		/* strip extension */
    for (cp = p; *cp != '\0'; cp++)
      if (!isalnum(*cp))
	*cp = '_';
  }
  return [String str:p];
}

+ bindName
{
  STR s;
  static id name;

  if (name)
    return name;

  s = [[options srcfilename] str];
  if (!s)
    s = [[options inFile] str];
  if (!s)
    s = "stdin";
  return name = ocbindname(s);
}

+ objcbindName
{
  return [[String str:"_OBJCBIND_"] concatSTR:[[self bindName] str]];
}

+ moddescname
{
  /* before cplus support we just used to call this variable "_modDesc"
   * and made it static to each module.  static initializers don't
   * seem to work in C++ as in C, and because we need a reference
   * from _modDesc to class impl and from class impl to _modDesc, we just
   * make this variable have a unique global name.
   */

  id s = [self bindName];

  return [String sprintf:"%s_modDesc", [s str]];
}

+ allClasses
{
  id c = nil;

  if ([options onePfile]) {
    c = (classimpl) ? [[OrdCltn new] add : classimpl]:nil;
  } else {
    c = clsCltn;
  }

  return c;
}


+ globfuncallOn:stream
{
  int i, n;
  id all = [self allClasses];

  n = (all) ? [all size] : 0;
  for (i = 0; i < n; i++)
    [[all at:i] globfuncallOn:stream];

  return self;
}

+ entrypointOn:stream
{
  id bName = [self objcbindName];

  /* int selSize = [selCltn size]; */

  [stream newline];

  [stream echo:"struct modDescriptor "];

  /* for DLL's */
  [stream echo:[[options bind] str]];

  [stream echo:" *%s(void)\n{\n", [bName str]];
  [stream echo:"\tselTransTbl = _selTransTbl;\n"];
  if ([options fwd])
    [stream echo:"\tfwdTransTbl = _fwdTransTbl;\n"];

  /* -noShared case, where we can't have global data shared across
   * modules
   */

  if ([options sharedData] == NO) {
    [self globfuncallOn:stream];
  }
  [stream echo:"\treturn &%s;\n}\n", [[self moddescname] str]];

  return self;
}

/*****************************************************************************
 *
 * struct useDescriptor {
 *	int processed;
 *	struct useDescriptor *next;
 *	struct useDescriptor ***uses;
 *	struct modDescriptor *(*bind)();
 * };
 *
 ****************************************************************************/

static id usesEntries, definesEntries, staticEntries;

+ staticEntry:aString
{
  if (!staticEntries)
    staticEntries = [Set new];
  if ([options debugInfo])
    printf("static use %s\n", [aString str]);
  assert([aString isKindOf:[String class]]);
  [staticEntries add:aString];
  return self;
}

+ usesEntry:aString
{
  if (!usesEntries)
    usesEntries = [Set new];
  if ([options debugInfo])
    printf("use %s\n", [aString str]);
  assert([aString isKindOf:[String class]]);
  if (![staticEntries contains:aString])
    [usesEntries add:aString];
  return self;
}

+ definesEntry:aString
{
  if (!definesEntries)
    definesEntries = [Set new];
  if ([options debugInfo])
    printf("define %s\n", [aString str]);
  assert([aString isKindOf:[String class]]);
  if (![staticEntries contains:aString])
    [definesEntries add:aString];
  return self;
}

+ (BOOL) definesMain
{
  return [definesEntries contains:[options mainFunction]];
}

+ usesOn:stream
{
  id aSeq, entry;

  if ([options comments])
    [stream echo:"\n/* uses */\n"];

  aSeq = [usesEntries eachElement];
  while ((entry = [aSeq next])) {
    STR s = [entry str];

    /* we do NOT define the "uses" entries as "extern"

     * this is a somewhat less understood feature of K&R C 
     * and most ANSI C compilers also implement it (under
     * the name of 'relaxed ANSI' or so)
     *
     * other compilers, such as the SGI IRIX, have a flag
     * for it (such as -common).
     *
     * the idea is that uninitialized globals are supposed
     * to be placed into the 'COMMON' segment per tranlation
     * unit, instead of immediately placing the symbol in the
     * BSS segment.  then when linking, multiple COMMON
     * symbols are allowed (and merged) and if the symbol's
     * also in the DATA segment, it gets a value.  Otherwise
     * it goes into the BSS segment (i.e. it will be 0).
     *
     * we define the entry in multiple modules
     * (e.g. OCU_MyObject) and to use the definition that has
     * an initializer (that is the "defines" entry).
     *
     * for some functions (e.g. OCU_printf) there will be no
     * define, but we will not get an unresolved symbol...
     *
     * it is necessary to generate OCU_funcs because we might
     * have a module, to which our only entry point is a funccall,
     * and that itself uses Objective C classes that need init...
     */

    [stream echo:"struct useDescriptor *OCU_%s;\n", s];
  }
  aSeq = [aSeq free];

  return self;
}

+ usecontrolOn:stream
{
  id aSeq, entry;
  id bName = [self objcbindName];

  [stream newline];
  [stream echo:"static struct useDescriptor **_useControl[] = {\n"];

  aSeq = [usesEntries eachElement];
  while ((entry = [aSeq next])) {
    STR s = [entry str];

    [stream echo:"&OCU_%s,\n", s];
  }
  aSeq = [aSeq free];

  [stream echo:"0};\n"];

  [stream newline];

  [stream echo:"static struct useDescriptor _useDesc = {\n"];
  [stream echo:"\t0,\n"];
  [stream echo:"\t0,\n"];
  [stream echo:"\t_useControl"];

  /* stes apr/22/1997 - removed if (selSize) test for case where
   * _OBJCBIND function was wrongly set to 0 ("super" example in testdir)
   * we can always use the bind function (since it's defined anyway) and
   * does itself a if (selSize) test.
   */

  [stream echo:",\n\t%s", [bName str]];

  [stream echo:"\n};\n"];

  return self;
}

+ definesOn:stream
{
  id aSeq, entry;
  id mf = [options mainFunction];

  if ([options comments])
    [stream echo:"\n/* defines */\n"];
  aSeq = [definesEntries eachElement];
  while ((entry = [aSeq next])) {
    STR f, s = [entry str];

    f = "struct useDescriptor *OCU_%s = &_useDesc;\n";
    [stream echo:f, s];
  }
  aSeq = [aSeq free];

  /* objcrt references OCU_main but our main function can be OCU_bar */

  if ([self definesMain] && (![mf isEqualSTR:"main"])) {
    [stream echo:"\nstruct useDescriptor *OCU_main=&_useDesc;\n"];
  }
  return self;
}

+ usedescOn:stream
{
  [stream newline];

  [stream echo:"struct useDescriptor {\n"];
  [stream echo:"\tint processed;\n"];
  [stream echo:"\tstruct useDescriptor *next;\n"];
  [stream echo:"\tstruct useDescriptor ***uses;\n"];
  [stream echo:"\tstruct modDescriptor *(*bind)();\n"];
  [stream echo:"};\n"];

  [self usesOn:stream];
  [self usecontrolOn:stream];
  [self definesOn:stream];

  /* turn on auto-initialization code by defining the postlink
   * entry point as zero.
   * objcrt initialization has an extern to this symbol 
   */

  if ([self definesMain]) {
    [stream echo:"\nstruct modEntry *_objcModules = 0;\n"];
  }
  return self;
}

+ postlinkmarkOn:stream
{
  id bName = [self bindName];

  /* the postLink version doesn't need the OCU entries
   * however, we want to 'mark' the .o file as being
   * compiled with -postlink, so that when linking
   * against a library that was compiled with -postlink
   * we could automatically detect this (and use postlink).
   */

  [stream newline];
  [stream echo:"int _OBJCPOSTLINK_%s = 1;\n", [bName str]];

  /* turn off auto-initialization code
   * objcrt auto-initialization has an extern to this symbol
   */

  if ([self definesMain]) {
    [stream echo:"\nstruct useDescriptor *OCU_main = 0;\n"];
  }
  return self;
}

+ otbmark:stream
{
  id bName = [self bindName];

  [stream newline];
  [stream echo:"int _OBJCOTB_%s = 1;\n", [bName str]];
  return self;
}

+ epilogueOn:stream
{
  if ([options comments])
    [stream echo:"\n\n/* objc epilogue */\n"];

  if ([options fwd])
    [self fwdstubsOn:stream];

  [self seltranstblOn:stream];
  if ([options fwd])
    [self fwdtranstblOn:stream];

  if (![options onePfile])
    [self modclslstOn:stream];
  [self moddescOn:stream];
  [self entrypointOn:stream];

  if ([options postLink]) {
    [self postlinkmarkOn:stream];
  } else {
    [self usedescOn:stream];
  }

  if ([options otb])
    [self otbmark:stream];

  return self;
}

@end

