
/*
 * Portable Object Compiler (c) 1997,98.  All Rights Reserved.
 * $Id: msgxpr.m,v 1.22 1998/10/06 15:23:45 stes Exp $
 */

/*
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "msgxpr.h"
#include "linetag.h"
#include "options.h"
#include "output.h"
#include "classdef.h"
#include "methinfo.h"
#include "identif.h"
#include "methimpl.h"
#include "ocstring.h"
#include "ordcltn.h"
#include "keywdecl.h"
#include "methdecl.h"
#include "compstat.h"
#include <assert.h>

@implementation MessageExpression

static int varCounter = 0;
static int cachecount = 0;

- processReceiver:aReceiver
{
  receiver = aReceiver;

  /* count of the temporary variable that will hold the receiver */
  tmpVar = [String sprintf:"objcT%i", varCounter++];

  /* name of the inlineCache (if we're doing this) */
  if ([options inlineCache]) {
    inlineCache = [String sprintf:"objcIC%i", cachecount++];
  }
  rcvrSuper = [receiver isEqualSTR:"super"];

  if (rcvrSuper) {
    id cName = [ClassDefinition currentName];

    if (!cName)
      fatalError("message to super outside @implementation");
  }
  /* idea here is that linetag associated to the entire msgxpr
   * is equal to linetag of left '['; this tag must be emitted before
   * any generated code
   */

  linetag = [[subnodes at:0] lineTag];

  return self;
}

- msgargs:methdecl
{
  methodMsg = [methdecl method];
  methodName = [methodMsg fullname];
  methodProto = [MethodInfo findproto:methodMsg];

  [methodMsg checkwith];	/* +with: gadget */

  return self;
}

- dispatch:rcvr method:minfo
{
  receiver = rcvr;
  tmpVar = rcvr;
  methodName = [minfo fullname];
  methodProto = minfo;
  methodMsg = minfo;
  dispatch = YES;
  return self;
}

- preventfwd:(BOOL)flag
{
  [methodProto preventfwd:flag];
  return self;
}

- (BOOL) canforward
{
  return [methodProto canforward];
}

- (BOOL) isselptr
{
  return [methodProto isselptr];
}

- methodProto
{
  return methodProto;
}

/* messages are 'equal' for *forwarding* if they have the same type
 * of argument and return value
 * this is to speed up module's fwdOffset: method
 */

- (unsigned) hash
{
  return [methodProto typehash];
}

- (BOOL) isEqual:msgxpr
{
  return (self == msgxpr) ? YES : [methodProto istypeEqual:[msgxpr methodProto]];
}

- grepTmpvars:aStmt
{
  /* add this tmp var to the list to be declared by the statement */
  [aStmt addTmpvar:tmpVar];

  /* handle subnodes (such as the receiver) */
  return [super grepTmpvars:aStmt];
}

- grepCaches:aStmt
{
  assert([options inlineCache]);

  /* add this tmp var to the list to be declared by the statement */
  [aStmt addCache:inlineCache];

  /* handle subnodes (such as the receiver) */
  return [super grepCaches:aStmt];
}


/* utilities for creating msgxprs */

- receiver:rcvr keyw:kw
{
  id m;

  [self processReceiver:rcvr];
  m = [MethodDeclarator new];
  [m setSelector:kw];
  [self msgargs:m];
  return self;
}

+ receiver:rcvr keyw:kw
{
  return [[self new] receiver:rcvr keyw:kw];
}

- receiver:rcvr keyw:kw1 arg:a1
{
  id m, k;

  [self processReceiver:rcvr];
  m = [MethodDeclarator new];
  k = [KeyWordDeclarator new];
  [k setName:kw1];
  [k setArgument:a1];
  [m setSelector:[[Node new] add:k]];
  [self msgargs:m];
  return self;
}

+ receiver:rcvr keyw:kw1 arg:a1
{
  return [[self new] receiver:rcvr keyw:kw1 arg:a1];
}

- receiver:rcvr keyw:kw1 arg:a1 keyw:kw2 arg:a2
{
  id m, k1, k2;

  [self processReceiver:rcvr];
  m = [MethodDeclarator new];
  k1 = [KeyWordDeclarator new];
  [k1 setName:kw1];
  [k1 setArgument:a1];
  k2 = [KeyWordDeclarator new];
  [k2 setName:kw2];
  [k2 setArgument:a2];
  [m setSelector:[[[Node new] add:k1] add:k2]];
  [self msgargs:m];
  return self;
}

+ receiver:rcvr keyw:kw1 arg:a1 keyw:kw2 arg:a2
{
  return [[self new] receiver:rcvr keyw:kw1 arg:a1 keyw:kw2 arg:a2];
}


- genOn:stream
{
  [stream newline];
  [linetag outputOn:stream];

  if (rcvrSuper) {
    [self supermsgOn:stream];
  } else {
    [self msgOn:stream];
  }

  return self;
}

- assignrcvrOn:stream
{
  STR tmpvarname = [tmpVar str];

  [stream echo:"%s = ", tmpvarname];
  [LineTag forceTag];
  [receiver genOn:stream];
  return self;
}

- assignsuperOn:stream
{
  STR tmpvarname = [tmpVar str];
  id c = [ClassDefinition current];

  if (!c)
    return self;

  /* the meaning of "super" depends on whether we're in a factory */
  /* or instance method implementation                          */

  /* PS: this could be simplified if we'd always use the Windows DLL
   * case, where we access the meta or class struct via a function
   * stes 3/98 made that change for -dynamic and -otb,
   * can only be slightly slower.  for the plain old case, keep the
   * original as to not confuse people... (runtime libs need recompiled)
   */

  [stream echo:"%s = ", tmpvarname];

  if ([currimpl isFactory]) {
    [c metasuperOn:stream];
  } else {
    [c superOn:stream];
  }

  return self;
}

- genimpOn:stream inSuper:(BOOL)inSuper
{
  STR funcname;
  STR tmp = [tmpVar str];

  /* regular, non-optimized messenger call */

  if (![options fwd]) {
    funcname = (inSuper) ? "_impSuper" : "_imp";
    [stream echo:"%s(%s,", funcname, tmp];
    if (dispatch) {
      [stream echo:"_cmd"];
    } else {
      [stream echoSel:methodName];
    }
    [stream echo:")"];
  } else {
    funcname = (inSuper) ? "fwdimpSuper" : "fwdimp";
    [stream echo:"%s(%s,", funcname, tmp];
    if (dispatch) {
      [stream echo:"_cmd"];
    } else {
      [stream echoSel:methodName];
    }
    [stream echo:","];
    [stream echoFwd:self];
    [stream echo:")"];
  }

  return self;
}

- gencacheOn:stream inSuper:(BOOL)inSuper
{
  STR getclass;
  STR tmp = [tmpVar str];
  STR cache = [inlineCache str];

  /* this is the optimized messenger call */

  /* check for messages to nil */
  [stream echo:"(%s)?(", tmp];

  /* check for class pointer equal */
  [stream echo:"("];
  if (![options otb]) {
    getclass = (inSuper) ? "%s" : "%s->isa";
  } else {
    getclass = (inSuper) ? "%s" : "%s->ptr->isa";
  }
  [stream echo:getclass, tmp];
  [stream echo:" == %s.cls)?%s.imp:", cache, cache];

  [stream echo:"("];
  /* cache failure, do regular lookup */
  [stream echo:"%s.cls=", cache];
  [stream echo:getclass, tmp];
  [stream echo:",%s.imp=", cache];
  [self genimpOn:stream inSuper:inSuper];
  [stream echo:")"];

  [stream echo:"):_nilHandler"];
  return self;
}

- gencallOn:stream inSuper:(BOOL)inSuper
{
  /* call messenger, cast return value, and dereference */

  [stream echo:"(*("];
  [methodProto castOn:stream];
  [stream echo:")"];

  if ([options inlineCache] && dispatch == NO) {
    [stream echo:"("];
    [self gencacheOn:stream inSuper:inSuper];
    [stream echo:")"];
  } else {
    [self genimpOn:stream inSuper:inSuper];
  }

  [stream echo:")"];

  return self;
}

- msgargsOn:stream inSuper:(BOOL)inSuper
{
  /* temp var for receiver & selector _cmd */

  if (inSuper) {
    /* can be inside a block ! then we output scope->self */
    [Identifier selfidOn:stream];
    [stream echo:","];
  } else {
    STR tmp = [tmpVar str];

    [stream echo:"%s,", tmp];
  }

  if (dispatch) {
    [stream echo:"_cmd"];
  } else {
    [stream echoSel:methodName];
  }

  [LineTag forceTag];
  [methodMsg genargsOn:stream];	/* begins with comma */

  return self;
}

- genmsgOn:stream inSuper:(BOOL)inSuper
{
  [self gencallOn:stream inSuper:inSuper];
  [stream echo:"("];
  [self msgargsOn:stream inSuper:inSuper];
  [stream echo:")"];

  return self;
}

- msgOn:stream
{
  [stream echo:"("];
  [self assignrcvrOn:stream];
  [stream echo:","];
  [self genmsgOn:stream inSuper:NO];
  [stream echo:")"];

  [LineTag forceTag];
  return self;
}

- supermsgOn:stream
{
  [stream echo:"("];
  [self assignsuperOn:stream];
  [stream echo:","];
  [self genmsgOn:stream inSuper:YES];
  [stream echo:")"];

  [LineTag forceTag];
  return self;
}

@end

