
/*
 * Portable Object Compiler (c) 1997.  All Rights Reserved.
 * $Id: node.m,v 1.25 1998/10/06 15:23:45 stes Exp $
 */

/*
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "node.h"
#include "ordcltn.h"
#include "linetag.h"
#include "output.h"
#include "options.h"

@implementation Node
/*****************************************************************************
 *
 * Errors
 *
 ****************************************************************************/

int exitStatus = 0;
id allTypenames;
id allClassnames;

/* stes - mar 26, 1997
 * change format from
 * objc: syntax error "translate" at test.m:358
 * to
 * test.m:358: [objc1] syntax error at "translate"
 * which is the standard format for emacs and vi
 */

void 
warnError(STR msg)
{
  STR name = [LineTag curfilename];
  int lineNo = [LineTag curlineno];
  STR program = ([options cplus]) ? "objcpls1" : "objc1";

  fprintf(stderr, "%s:%d: warning (%s): %s\n", name, lineNo, program, msg);
}

void 
fatalError(STR msg)
{
  STR name = [LineTag curfilename];
  int lineNo = [LineTag curlineno];
  STR program = ([options cplus]) ? "objcpls1" : "objc1";

  fprintf(stderr, "%s:%d: error (%s): %s\n", name, lineNo, program, msg);
  exitStatus = 1;
}

/*****************************************************************************
 *
 * Creation
 *
 ****************************************************************************/


- node
{
  subnodes = nil;		/* created only when needed */
  return self;
}

+ new
{
  return [[super new] node];
}

- dodeepCopy
{
  subnodes = [subnodes deepCopy];
  return self;
}

- deepCopy
{
  /* not all super's have a deepCopy */
  return [[super copy] dodeepCopy];
}

- free
{
  if (subnodes)
    subnodes = [[subnodes freeContents] free];
  return [super free];
}


- subnodes
{
  return subnodes;
}

- (BOOL) isEqual:aNode
{
  if (isa != [aNode class])
    return NO;

  if (subnodes) {
    int n = [subnodes size];
    id aSubnodes = [aNode subnodes];

    if (aSubnodes == nil || n != [aSubnodes size])
      return NO;
    while (n--)
      if (![[subnodes at:n] isEqual:[aSubnodes at:n]])
	return NO;
  }
  return YES;
}

- (BOOL) isEqualSTR:(STR)s
{
  if (subnodes != nil && [subnodes size] == 1) {
    return [[subnodes at:0] isEqualSTR:s];
  } else {
    return NO;
  }
}

- (int) asInt
{
  if (subnodes != nil && [subnodes size] == 1) {
    return [[subnodes at:0] asInt];
  } else {
    return 0;
  }
}

/*****************************************************************************
 *
 * Do Semantical Analysis.
 * |process| can be overwritten to do things like typeChecking, defining
 * types etc.
 *
 ****************************************************************************/

- process
{
  return self;
}


/*****************************************************************************
 *
 * Dealing with "typedef".
 * Our scan.l discriminates IDENTIFIER_TOKEN from TYPENAME_TOKEN, by table
 * lookup in a global "allTypenames".
 *
 ****************************************************************************/


- (STR) str
{
  if (subnodes && [subnodes size] == 1) {
    return [[subnodes at:0] str];
  }
  return NULL;
}

- (BOOL) isPtr
{
  if (subnodes && [subnodes size] == 1) {
    return [[subnodes at:0] isPtr];
  }
  return NO;
}


- (BOOL) isId
{
  if (subnodes && [subnodes size] == 1) {
    return [[subnodes at:0] isId];
  }
  return NO;
}

- (BOOL) isVoid
{
  if (subnodes && [subnodes size] == 1) {
    return [[subnodes at:0] isVoid];
  }
  return NO;
}

- (BOOL) containsTypedef
{
  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++) {
      if ([[subnodes at:i] containsTypedef])
	return YES;
    }
  }
  return NO;
}

- deftypename
{
  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++)
      [[subnodes at:i] deftypename];
  }
  return self;
}

- markdecl:(BOOL)flag
{
  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++)
      [[subnodes at:i] markdecl:flag];
  }
  return self;
}

/*****************************************************************************
 *
 * Dealing with temporary variables & inline Caches
 * The messenger needs temporary variables for receivers.
 * Methods to find out whether there's message expressions.
 * These are being sent by the compound statement (with itself as argument).
 *
 ****************************************************************************/


- grepTmpvars:aStmt
{
  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++) {
      [[subnodes at:i] grepTmpvars:aStmt];
    }
  }
  return self;
}

- grepBlocks:aStmt
{
  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++) {
      [[subnodes at:i] grepBlocks:aStmt];
    }
  }
  return self;
}

- grepTopblks:aStmt
{
  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++) {
      [[subnodes at:i] grepTopblks:aStmt];
    }
  }
  return self;
}

- grepCaches:aStmt
{
  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++) {
      [[subnodes at:i] grepCaches:aStmt];
    }
  }
  return self;
}

- processBlocks
{
  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++) {
      [[subnodes at:i] processBlocks];
    }
  }
  return self;
}

- genblockdefsOn:stream
{
  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++) {
      [[subnodes at:i] genblockdefsOn:stream];
    }
  }
  return self;
}

- grepScope:aBlock
{
  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++) {
      [[subnodes at:i] grepScope:aBlock];
    }
  }
  return self;
}

/*****************************************************************************
 *
 * Dealing with "main()".
 * We want to insert an extra function call (to initialize the runtime).
 *
 ****************************************************************************/


- findIdentifier
{
  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++) {
      id ident = [[subnodes at:i] findIdentifier];

      if (ident)
	return ident;
    }
  }
  return nil;
}

- (BOOL) containsStatic
{
  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++) {
      if ([[subnodes at:i] containsStatic])
	return YES;
    }
  }
  return NO;
}


/*****************************************************************************
 *
 * Addding subnodes
 *
 ****************************************************************************/


- add:aNode
{
  if (!subnodes)
    subnodes = [OrdCltn new];
  /* allow nil nodes for "empty" rules */
  if (aNode)
    [subnodes add:aNode];
  return self;
}


/*****************************************************************************
 *
 * Assembling instance variables for @interface
 *
 ****************************************************************************/

- grepIvars:aSet addItype:bNode to:bSet
{
  /* make a copy so that int *x, y; will store type
   * int* for x, and type int for y
   */

  id aCopy = [bNode deepCopy];

  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++) {
      [[subnodes at:i] grepIvars:aSet addItype:aCopy to:bSet];
    }
  }
  return self;
}

- grepIvars:aSet initvars:iv addItype:bNode to:bSet
{
  id aCopy = [bNode deepCopy];

  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++) {
      [[subnodes at:i] grepIvars:aSet initvars:iv addItype:aCopy to:bSet];
    }
  }
  return self;
}


/*****************************************************************************
 *
 * Assembling Names, Arguments etc. for Method
 *
 ****************************************************************************/

- forMethod:aMethod
{
  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++) {
      [[subnodes at:i] forMethod:aMethod];
    }
  }
  return self;
}

- argsstmt:statements
{
  if (subnodes)
    [subnodes elementsPerform:_cmd with:statements];
  return self;
}


/*****************************************************************************
 *
 * Output code
 *
 ****************************************************************************/


- outputOn:stream
{
  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++) {
      [[subnodes at:i] outputOn:stream];
    }
  }
  return self;
}

- taglessOn:stream
{
  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++) {
      [[subnodes at:i] taglessOn:stream];
    }
  }
  return self;
}

- genfieldOn:stream
{
  return [self genOn:stream];
}

- funheadOn:stream
{
  return [self genOn:stream];
}

- genOn:stream
{
  if (subnodes) {
    int i, n = [subnodes size];

    for (i = 0; i < n; i++) {
      [[subnodes at:i] genOn:stream];
    }
  } else {
    [self outputOn:stream];
  }

  return self;
}

- printOn:(IOD)f
{
  id stream = [OutputStream newFile:f];

  [self outputOn:stream];
  return self;
}

@end

