
/*
 * Portable Object Compiler (c) 1997.  All Rights Reserved.
 * $Id: options.m,v 1.19 1998/10/06 15:23:45 stes Exp $
 */

/*
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "options.h"
#include "ocstring.h"
#include <stdlib.h>		/* getenv */

id options;

@implementation Options

/*****************************************************************************
 *
 * Creation
 *
 ****************************************************************************/


/* stes 12/97 - unfortunate choice of me was to use the short form
 * # %d %s as default (as in gcc) ANSI C uses #line, and so most
 * compilers had to be invoked with -longlinetag, which is silly
 */
static char *defaultformat = "#line %d \"%s\"";

+ new
{
  if (options)
    return options;

  options = [super new];
  [options quiet:NO];
  [options outputCode:YES];
  [options genCode:YES];
  [options otb:NO];		/* by default, no Object Tables */
  [options fwd:YES];		/* use fwdimp() ? */
  [options selptr:YES];		/* use selptrfwd ? */
  [options version:NO];
  [options inFile:nil /* stdin */ ];
  [options setOutFile:nil /* stdout */ ];
  /* nil is understood by the LineTag class as "use inFile" */
  [options srcfilename:nil];
  [options tagformat:[String str:defaultformat]];
  [options initCall:nil];	/* no default - always set at driver level */
  [options mainfun:[String str:"main"]];
  [options seltranslation:YES];
  [options inlineCache:NO];
  [options linetags:YES];
  /* default is to do auto-initialization and to use regular "ld" */
  [options postlink:NO];
  /* default is to assume shared data is available */
  [options shareddata:YES];
  [options refBind:NO];		/* whether to force a reference to BIND */
  [options checkBind:NO];	/* used in selector output */
  [options warnclasstype:YES];	/* warn by default */
  [options warntypeconflict:YES];	/* warn by default */
  [options warnundefined:YES];	/* warn by default */
  [options warnnotfound:YES];
  [options warnlocalnst:YES];	/* local hides instance warn */
  [options warnlex:NO];		/* 0xff problem */
  [options buffered:YES];	/* use block buffering */
  [options comments:NO];	/* so that output file can have .i extension */
  [options msdos:NO];		/* by default, no MS-DOS keywords */
  [options gnu:NO];		/* by default, no GNU keywords */
  [options setCplus:NO];	/* by default, not cplusplus mode */
  [options debugInfo:NO];	/* by default, dont' output debug info */
  [options ppi:NO];		/* by default, no PPI syntax */
  [options ibmvac:NO];		/* by default, no Visual Age keywords */

  /* default separator for this platform (at compile time) */
  [options pathsep:OBJCRT_DEFAULT_PATHSEPC];

  /* default is to allow more than one implementation per class */
  [options onePfile:NO];

  /* DEC alpha defines a asm() function... */
  [options asmKeyword:YES];

  /* SGI IRIX __long_long keyword support */
  [options llkeyword:YES];

  /* allow Objective-C blocks extensions by default */
  [options blocks:YES];

  /* link the AsciiFiler class into program */
  [options filerCode:YES];

  /* do not by default emit __declspec(export) for OBJCBIND */
  [options bind:[String new]];

  /* no max by default */
  [options linemax:0];

  /* objcrt control */
  [options cache:YES];
  [options nilRcvr:YES];

  /* by default, allow self= */
  [options selfAssign:YES];

  /* purists can set this to NO and use +addMethodsTo:  */
  [options categories:YES];

  return options;
}


- linemax:(int)flag
{
  lineMax = flag;
  return self;
}

- (int) lineMax
{
  return lineMax;
}

- categories:(BOOL)flag
{
  cats = flag;
  return self;
}

- (BOOL) cats
{
  return cats;
}

- otb:(BOOL)flag
{
  otb = flag;
  return self;
}

- (BOOL) otb
{
  return otb;
}

- fwd:(BOOL)flag
{
  fwd = flag;
  return self;
}

- (BOOL) fwd
{
  return fwd;
}

- selptr:(BOOL)flag
{
  selptr = flag;
  return self;
}

- (BOOL) selptr
{
  return selptr;
}

- selfAssign:(BOOL)flag
{
  selfAssign = flag;
  return self;
}

- (BOOL) selfAssign
{
  return selfAssign;
}

- refBind:(BOOL)flag
{
  /* whether to force a reference to the OBJCBIND function
   * needed on Metrowerks compiler
   */
  refBind = flag;
  return self;
}

- (BOOL) refBind
{
  return refBind;
}

- nilRcvr:(BOOL)flag
{
  nilRcvr = flag;
  return self;
}

- (BOOL) nilRcvr
{
  return nilRcvr;
}

- cache:(BOOL)flag
{
  cache = flag;
  return self;
}

- (BOOL) cache
{
  return cache;
}

- bind:arg
{
  bind = arg;
  return self;
}

- bind
{
  return bind;
}

- filerCode:(BOOL)flag
{
  filerCode = flag;
  return self;
}

- (BOOL) filerCode
{
  return filerCode;
}

- blocks:(BOOL)flag
{
  blocks = flag;
  return self;
}

- (BOOL) blocks
{
  return blocks;
}

- llkeyword:(BOOL)flag
{
  llKeyword = flag;
  return self;
}

- (BOOL) llKeyword
{
  return llKeyword;
}

- asmKeyword:(BOOL)flag
{
  asmKeyword = flag;
  return self;
}

- (BOOL) asmKeyword
{
  return asmKeyword;
}

- onePfile:(BOOL)flag
{
  onePfile = flag;
  return self;
}

- (BOOL) onePfile
{
  return onePfile;
}

- quiet:(BOOL)flag
{
  quiet = flag;
  return self;
}

- (BOOL) quiet
{
  return quiet;
}

- pathsep:(STR)aSep
{
  pathSeparator = (STR) getenv("OBJCPATHSEP");
  pathSeparator = (pathSeparator) ? pathSeparator : aSep;
  return self;
}

- (STR) pathSeparator
{
  return pathSeparator;
}

- ppi:(BOOL)flag
{
  ppi = flag;
  return self;
}

- (BOOL) ppi
{
  return ppi;
}

- debugInfo:(BOOL)flag
{
  debugInfo = flag;
  return self;
}

- (BOOL) debugInfo
{
  return debugInfo;
}

- setCplus:(BOOL)flag
{
  cplus = flag;
  return self;
}

- (BOOL) cplus
{
  return cplus;
}

- msdos:(BOOL)flag
{
  msdos = flag;
  return self;
}

- (BOOL) msdos
{
  return msdos;
}

- watcom:(BOOL)flag
{
  watcom = flag;
  return self;
}

- (BOOL) watcom
{
  return watcom;
}

- ibmvac:(BOOL)flag
{
  ibmvac = flag;
  return self;
}

- (BOOL) ibmvac
{
  return ibmvac;
}

- gnu:(BOOL)flag
{
  gnu = flag;
  return self;
}

- (BOOL) gnu
{
  return gnu;
}

- comments:(BOOL)flag
{
  comments = flag;
  return self;
}

- (BOOL) comments
{
  return comments;
}

- outputCode:(BOOL)flag
{
  outputCode = flag;
  return self;
}

- (BOOL) outputCode
{
  return outputCode;
}

- seltranslation:(BOOL)flag
{
  seltranslation = flag;
  return self;
}

- (BOOL) seltranslation
{
  return seltranslation;
}

- genCode:(BOOL)flag
{
  genCode = flag;
  return self;
}

- (BOOL) genCode
{
  return outputCode;
}

- buffered:(BOOL)flag
{
  isBuffered = flag;
  return self;
}

- (BOOL) isBuffered
{
  return isBuffered;
}

- linetags:(BOOL)flag
{
  lineTags = flag;
  return self;
}

- (BOOL) lineTags
{
  return lineTags;
}

- shareddata:(BOOL)flag
{
  sharedData = flag;
  return self;
}

- (BOOL) sharedData
{
  return sharedData;
}

- version:(BOOL)flag
{
  version = flag;
  return self;
}

- (BOOL) version
{
  return version;
}

- inlineCache:(BOOL)flag
{
  inlineCache = flag;
  return self;
}

- (BOOL) inlineCache
{
  return inlineCache;
}

- warnlex:(BOOL)flag
{
  warnLex = flag;
  return self;
}

- (BOOL) warnLex
{
  return warnLex;
}

- warnclasstype:(BOOL)flag
{
  warnclasstype = flag;
  return self;
}

- (BOOL) warnclasstype
{
  return warnclasstype;
}

- warntypeconflict:(BOOL)flag
{
  warntypeconflict = flag;
  return self;
}

- (BOOL) warntypeconflict
{
  return warntypeconflict;
}

- warnnotfound:(BOOL)flag
{
  warnnotfound = flag;
  return self;
}

- (BOOL) warnnotfound
{
  return warnnotfound;
}

- warnlocalnst:(BOOL)flag
{
  warnlocalnst = flag;
  return self;
}

- (BOOL) warnlocalnst
{
  return warnlocalnst;
}

- warnundefined:(BOOL)flag
{
  warnundefined = flag;
  return self;
}

- (BOOL) warnundefined
{
  return warnundefined;
}

- checkBind:(BOOL)flag
{
  checkbind = flag;
  return self;
}

- (BOOL) checkbind
{
  return checkbind;
}

- postlink:(BOOL)flag
{
  postLink = flag;
  return self;
}

- (BOOL) postLink
{
  return postLink;
}

- inFile:aString
{
  inFile = aString;
  return self;
}

- inFile
{
  return inFile;
}

- srcfilename:aString
{
  srcfilename = aString;
  return self;
}

- srcfilename
{
  return srcfilename;
}

- setOutFile:aString
{
  outFile = aString;
  return self;
}

- outFile
{
  return outFile;
}

- tagformat:aString
{
  tagformat = aString;
  return self;
}

- tagformat
{
  return tagformat;
}

- initCall:aString
{
  initCall = aString;
  return self;
}

- initCall
{
  return initCall;
}

- mainfun:aString
{
  mainFunction = aString;
  return self;
}

- mainFunction
{
  return mainFunction;
}


@end

