
/*
 * Portable Object Compiler (c) 1997.  All Rights Reserved.
 * $Id: output.m,v 1.17 1998/10/06 15:23:45 stes Exp $
 */

/*
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "output.h"
#include "options.h"
#include "ocstring.h"
#include "methinfo.h"
#include "module.h"
#include <stdlib.h>

@implementation OutputStream
/*****************************************************************************
 *
 * Utility function
 *
 ****************************************************************************/

/*
 * Opening and closing a file.  We're not dealing with wrapping around
 * files (include).  This is because we're processing cpp output.
 */

FILE * openFile(STR name, STR modfs)
{
  FILE *f;

  if ((f = fopen(name, modfs)) == NULL) {
    fprintf(stderr, "objc1: Unable to open %s\n", name);
    exit(1);
  }
  return f;
}

/*****************************************************************************
 *
 * Creation
 *
 ****************************************************************************/


- file:(FILE *)f
{
  file = f;
  return self;
}

+ new
{
  return [self newFile:stdout];
}

+ str:(STR)s
{
  return [self newFile:openFile(s, "w")];
}

+ newFile:(FILE *)f
{
  /* stes - feb 23,1997
   * the default is to use block buffered output
   * but for debugging unbuffered output is very useful
   */

  if ([options isBuffered] == NO)
    setbuf(f, NULL);

  return [[super new] file:f];
}

- close
{
  fclose(file);
  return self;
}

/*****************************************************************************
 *
 * Output methods
 *
 ****************************************************************************/

- newline
{
  fprintf(file, "\n");
  return self;
}

- echo:(STR)format,...
{
  OC_VA_LIST ap;

  /* use OC macros for port to SunOS4 */

  OC_VA_START(ap, format);
  vfprintf(file, format, ap);
  OC_VA_END(ap);

  return self;
}

- echoId:anObject
{
  [anObject printOn:file];
  return self;
}

- echoFwd:message
{
  if ([message isselptr]) {
    [self echo:"selptrfwd"];	/* do most common cases special */
  } else {
    if ([message canforward]) {
      int offset = [Module fwdOffset:message];

      [self echo:"fwdTransTbl[%i]", offset];	/* general case */
    } else {
      [self echo:([options cplus]) ? "(id(*)(...))0" : "(id(*)())0"];
    }
  }
  return self;
}

- echoSel:aString
{
  if ([options seltranslation]) {
    BOOL check = [options checkbind];
    int offset = [Module selOffset:aString];

    if (check)
      [self echo:"(selTransTbl)?"];

    [self echo:"selTransTbl[%i", offset];
    if ([options comments])
      [self echo:" /* %s */", [aString str]];
    [self echo:"]"];

    if (check) {
      STR modName = [[Module bindName] str];

      [self echo:":objcrt_bindError(\"%s\")", modName];
    }
  } else {
    [self echo:"\"%s\"", [aString str]];
  }

  return self;
}

@end

