
/*
 * Portable Object Compiler (c) 1997.  All Rights Reserved.
 * $Id: token.m,v 1.16 1998/10/06 15:23:46 stes Exp $
 */

/*
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "token.h"
#include "linetag.h"
#include "ocstring.h"
#include "output.h"
#include "identif.h"

@implementation Token

/*****************************************************************************
 *
 * Creation
 *
 ****************************************************************************/


- setUp:(STR)aToken
{
  token = aToken;
  lineTag = [LineTag currentTag];
  return self;
}

- lineTag
{
  return lineTag;
}

+ newToken:(STR)aToken
{
  return [[super new] setUp:aToken];
}

+ str:(STR)aToken
{
  id s = [String str:aToken];

  return [[super new] setUp:[s str]];
}

- free
{
  return nil;
}



- outputOn:stream
{
  [LineTag outputOn:stream until:lineTag];
  /* watch out for % in tokenName */
  [stream echo:"%s", token];
  return self;
}

- taglessOn:stream
{
  /* watch out for % in tokenName */
  [stream echo:" %s", token];
  return self;
}

- tagOn:stream
{
  [LineTag outputOn:stream until:lineTag];
  return self;
}


- (STR) str
{
  return token;
}

- (BOOL) isEqual:aToken
{
  return (isa == [aToken class] && strcmp(token, [aToken str]) == 0);
}

- (BOOL) isEqualSTR:(STR)s
{
  return strcmp(token, s) == 0;
}

- (BOOL) containsTypedef
{
  return strcmp(token, "typedef") == 0;
}

- (BOOL) containsStatic
{
  /* this method returns YES for any keyword that should
   * prevent an OCU defining-entry to be generated for
   * a function definition that has the keyword
   * stes 12/97 : added inline for Debian header files
   */

  return (strcmp(token, "static") == 0)
      || (strcmp(token, "inline") == 0)
      || (strcmp(token, "__inline") == 0)
      || (strcmp(token, "__inline__") == 0);
}

- (BOOL) isPtr
{
  return strcmp(token, "id") == 0;
}

- (BOOL) isId
{
  return strcmp(token, "id") == 0;
}

- (BOOL) isVoid
{
  return strcmp(token, "void") == 0;
}

- findIdentifier
{
  return nil;
}

- grepIvars:aSet addItype:bNode to:bSet
{
  /*
   * stes 10/26 hack so that int* is type
   * we can add to 'bnode' since it's a copy for each identifier
   * of the type (e.g. int *x,y;)
   */
  if (strcmp(token, "*") == 0)
    [bNode add:[self copy]];
  return self;
}

- grepIvars:aSet initvars:iv addItype:bNode to:bSet
{
  return [self grepIvars:aSet addItype:bNode to:bSet];
}


- asIdentifier
{
  return [Identifier str:token];
}


#ifdef __PORTABLE_OBJC__

- fileOutOn:aFiler
{
  [super fileOutOn:aFiler];
  [aFiler fileOut:&token type:'*'];
  return self;
}

- fileInFrom:aFiler
{
  [super fileInFrom:aFiler];
  [aFiler fileIn:&token type:'*'];
  return self;
}

#endif				/* __PORTABLE_OBJC__ */

@end

